/*======================================================================================

	Program:		Range Imaging Toolkit (RITK)
	Module:			ritkCudaPitchedMemoryImportImageContainer.h
	Version:		$Revision: 3110 $
	Date:			$Date: 2012-04-18 17:32:46 +0200 (Mi, 18 Apr 2012) $

	Copyright (c) 2010-2011
	Pattern Recognition Lab, 
    University Erlangen-Nuremberg, Germany
    All rights reserved.

	This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See RITKCopyright.txt for more information.

	http://www5.informatik.uni-erlangen.de/research/software/ritk

========================================================================================*/

#ifndef CUDAPITCHEDMEMORYIMPORTIMAGECONTAINER_H__
#define CUDAPITCHEDMEMORYIMPORTIMAGECONTAINER_H__

#include "ritkCudaImportImageContainer.h"
#include <itkRGBAPixel.h>
#include <list>

namespace ritk
{

/**	@ingroup	Cuda
 *	@class		CudaPitchedMemoryImportImageContainer
 *	@author		Jakob Wasza
 *	@brief		CUDA pitched memory container
 *
 *	@details
 *	This class provides methods for CUDA pitched memory management.
 */
template<typename TElementIdentifier, typename TElement>
class CudaPitchedMemoryImportImageContainer : public CudaImportImageContainer<TElementIdentifier,TElement,2>
{
public:
	/**	@name	Std typedefs */
	//@{
	typedef CudaPitchedMemoryImportImageContainer						Self;
	typedef CudaImportImageContainer<TElementIdentifier,TElement,2>		Superclass;
	typedef itk::SmartPointer<Self>										Pointer;
	typedef itk::SmartPointer<const Self>								ConstPointer;
	//@}

	/// For convenience
	//@{
	typedef typename Superclass::GenericImportImageContainerType		GenericImportImageContainerType;
	typedef typename Superclass::ImportImageContainerType				ImportImageContainerType;
	//@}

	/**	@name	Template parameters */
	//@{
	typedef TElementIdentifier  ElementIdentifier;
	typedef TElement            Element;
	//@}

	//@}

	/// Object creation
	itkNewMacro(Self);

	/// RTTI
	itkTypeMacro(CudaPitchedMemoryImportImageContainer, CudaImportImageContainer);

	/**	@name	Host/device synchronization. Is delegated to the superclass */
	//@{
	virtual void SynchronizeDevice();
	virtual void SynchronizeHost();
	//@}

	/**	@name	Deep copy */
	//@{

	/**	@brief	Deep copy the buffer
	 *	@param	Container	The container providing the source
	 *
	 *	@details
	 *	Copy data from an itk::ImportImageContainer, however, this class' implementation only addresses copying device memory,
	 *	host memory copy will be performed by the baseclass that eventually also allocates memory (host AND device). \n
	 *	This special CUDA container will copy the CUDA memory if a CUDAImportImageContainer can be derived from the parameter.
	 *	This class' implementation will automatically determine and execute the proper copy scheme (i.e. Array -> Regular, ...).
	 *	Note that copying device memory will be performed if and only if the device modified time is equal or more up-to-date 
	 *	than the host memory. 
	 */
	virtual void DeepCopy(ImportImageContainerType *Container);


	/**	@brief	Deep copy from a container
	 *	@param	Container	The container providing the source
	 *
	 *	@details
	 *	This method will simply remove the const qualifier (const_cast) of the input argument and pass it to the non-const DeepCopy(ImportImageContainerType *) method.
	 *	Note that a non-const input is necessary due to the synchronization mechanism. However, it is granted, that the memory (host or device) that is up-to-data
	 *	will NOT be modified!
	 */
	virtual void DeepCopy(const ImportImageContainerType *Container);

	//@}

	/// Copy to a mapped resource
	virtual void CopyToMappedResource(void *d_ptr, size_t size) const;

	/**	@name	Pointer to the wrapped CUDA memory. Non-const version will modify the device MTime! */
	//@{
	Element *GetCudaMemoryPointer();
	const Element *GetCudaMemoryPointer() const;
	//@}

	/// Get the pitch of the wrapped memory. Note that the pitch is calculated from the 2-D size of the container. @sa SetSize
	size_t GetPitch() const;

protected:
	/// Constructor
	CudaPitchedMemoryImportImageContainer();
	/// Destructor
	virtual ~CudaPitchedMemoryImportImageContainer();

	/// Print
	virtual void PrintSelf(std::ostream& os, itk::Indent indent) const;

	/// Allocate the buffer. Is delegated to the superclass.
	virtual Element* AllocateElements(ElementIdentifier size) const;

	/// Free the buffer. Is delegated to the superclass
	virtual void DeallocateManagedMemory();

	/**	@name	Methods that have to be executed in the global CUDA context */
	//@{
	CudaConstMethodMacro(CudaPitchedMemoryImportImageContainer, AllocateCudaMemory);
	CudaMethodMacro(CudaPitchedMemoryImportImageContainer, DeallocateCudaMemory);
	CudaMethodMacro(CudaPitchedMemoryImportImageContainer, CopyDeviceToHost);
	CudaMethodMacro(CudaPitchedMemoryImportImageContainer, CopyHostToDevice);
	Superclass *m_TmpCudaContainer;
	CudaMethodMacro(CudaPitchedMemoryImportImageContainer, DeepCopyCudaMemory);
	//@}

protected:
	/// List holding the memory. This is due to the (de)allocate scheme in itk. Always delete the front and push new memory
	mutable std::list<std::pair<Element*,size_t> > m_CudaMemory;

private:
	/// purposely not implemented
	CudaPitchedMemoryImportImageContainer(const Self&); 
	/// purposely not implemented
	void operator=(const Self&);
};

/** @name Template instantiations */
//@{
typedef CudaPitchedMemoryImportImageContainer<unsigned long,float> CudaPitchedMemoryImportImageContainerF;
typedef CudaPitchedMemoryImportImageContainer<unsigned long,unsigned long> CudaPitchedMemoryImportImageContainerU;
typedef CudaPitchedMemoryImportImageContainer<unsigned long,itk::RGBAPixel<unsigned char> > CudaPitchedMemoryImportImageContainerUC4;
//@}

} // ritk


#endif // CUDAPITCHEDMEMORYIMPORTIMAGECONTAINER_H__
