/*======================================================================================

	Program:		Range Imaging Toolkit (RITK)
	Module:			ritkCudaRegularMemoryImportImageContainer.h
	Version:		$Revision: 3118 $
	Date:			$Date: 2012-04-19 10:09:28 +0200 (Do, 19 Apr 2012) $

	Copyright (c) 2010-2011
	Pattern Recognition Lab, 
    University Erlangen-Nuremberg, Germany
    All rights reserved.

	This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See RITKCopyright.txt for more information.

	http://www5.informatik.uni-erlangen.de/research/software/ritk

========================================================================================*/

#ifndef CUDAREGULARMEMORYIMPORTIMAGECONTAINER_H__
#define CUDAREGULARMEMORYIMPORTIMAGECONTAINER_H__

#include "ritkCudaImportImageContainer.h"

namespace ritk
{

/**	@ingroup	Cuda
 *	@class		CudaRegularMemoryImportImageContainer
 *	@author		Jakob Wasza
 *	@brief		CUDA regular memory container
 *
 *	@details
 *	This class provides methods for regular CUDA memory management.
 */
template<typename TElementIdentifier, typename TElement, unsigned int VDimension=1>
class CudaRegularMemoryImportImageContainer : public CudaImportImageContainer<TElementIdentifier,TElement,VDimension>
{
public:
	/**	@name	Std typedefs */
	//@{
	typedef CudaRegularMemoryImportImageContainer								Self;
	typedef CudaImportImageContainer<TElementIdentifier,TElement,VDimension>	Superclass;
	typedef itk::SmartPointer<Self>												Pointer;
	typedef itk::SmartPointer<const Self>										ConstPointer;
	//@}

	/// For convenience
	//@{
	typedef typename Superclass::GenericImportImageContainerType	GenericImportImageContainerType;
	typedef typename Superclass::ImportImageContainerType			ImportImageContainerType;
	//@}

	/**	@name	Template parameters */
	//@{
	typedef TElementIdentifier  ElementIdentifier;
	typedef TElement            Element;
	//@}

	//@}

	/** @name Dimension macros */
	//@{
	itkStaticConstMacro(Dimension, unsigned int, Superclass::Dimension);
	//@}

	/// Object creation
	itkNewMacro(Self);

	/// RTTI
	itkTypeMacro(CudaRegularMemoryImportImageContainer, CudaImportImageContainer);

	/**	@name	Host/device synchronization. Is delegated to the superclass */
	//@{
	virtual void SynchronizeDevice();
	virtual void SynchronizeHost();
	//@}

	/**	@name	Deep copy */
	//@{

	/**	@brief	Deep copy the buffer
	 *	@param	Container	The container providing the source
	 *
	 *	@details
	 *	Copy data from an itk::ImportImageContainer, however, this class' implementation only addresses copying device memory,
	 *	host memory copy will be performed by the baseclass that eventually also allocates memory (host AND device). \n
	 *	This special CUDA container will copy the CUDA memory if a CUDAImportImageContainer can be derived from the parameter.
	 *	This class' implementation will automatically determine and execute the proper copy scheme (i.e. Array -> Regular, ...).
	 *	Note that copying device memory will be performed if and only if the device modified time is equal or more up-to-date 
	 *	than the host memory. 
	 */
	virtual void DeepCopy(ImportImageContainerType *Container);


	/**	@brief	Deep copy from a container
	 *	@param	Container	The container providing the source
	 *
	 *	@details
	 *	This method will simply remove the const qualifier (const_cast) of the input argument and pass it to the non-const DeepCopy(ImportImageContainerType *) method.
	 *	Note that a non-const input is necessary due to the synchronization mechanism. However, it is granted, that the memory (host or device) that is up-to-data
	 *	will NOT be modified!
	 */
	virtual void DeepCopy(const ImportImageContainerType *Container);

	//@}

	/// Copy to a mapped resource
	virtual void CopyToMappedResource(void *d_ptr, size_t size) const;

	/**	@name	Pointer to the wrapped CUDA memory. Non-const version will modify the device MTime! */
	//@{
	Element *GetCudaMemoryPointer();
	const Element *GetCudaMemoryPointer() const;
	//@}


protected:
	/// Constructor
	CudaRegularMemoryImportImageContainer();
	/// Destructor
	virtual ~CudaRegularMemoryImportImageContainer();

	/// Print
	virtual void PrintSelf(std::ostream& os, itk::Indent indent) const;

	/// Allocate the buffer. Is delegated to the superclass.
	virtual Element* AllocateElements(ElementIdentifier size) const;

	/// Free the buffer. Is delegated to the superclass
	virtual void DeallocateManagedMemory();

	/**	@name	Methods that have to be executed in the global CUDA context */
	//@{
	CudaConstMethodMacro(CudaRegularMemoryImportImageContainer, AllocateCudaMemory);
	CudaMethodMacro(CudaRegularMemoryImportImageContainer, DeallocateCudaMemory);
	CudaMethodMacro(CudaRegularMemoryImportImageContainer, CopyDeviceToHost);
	CudaMethodMacro(CudaRegularMemoryImportImageContainer, CopyHostToDevice);
	Superclass *m_TmpCudaContainer;
	CudaMethodMacro(CudaRegularMemoryImportImageContainer, DeepCopyCudaMemory);
	//@}


protected:
	/// The actual CUDA memory 
	mutable std::list<Element*> m_CudaMemory;

private:
	/// purposely not implemented
	CudaRegularMemoryImportImageContainer(const Self&); 
	/// purposely not implemented
	void operator=(const Self&);
};

/** @name Template instantiations */
//@{
typedef CudaRegularMemoryImportImageContainer<unsigned long,float> CudaRegularMemoryImportImageContainerF;
typedef CudaRegularMemoryImportImageContainer<unsigned long,itk::Vector<float,3> > CudaRegularMemoryImportImageContainerF3;
typedef CudaRegularMemoryImportImageContainer<unsigned long,itk::Vector<unsigned char,3> > CudaRegularMemoryImportImageContainerUC3;
typedef CudaRegularMemoryImportImageContainer<unsigned long,unsigned long> CudaRegularMemoryImportImageContainerU;
typedef CudaRegularMemoryImportImageContainer<unsigned long,unsigned char> CudaRegularMemoryImportImageContainerUC;
//@}

} // ritk


#endif // CUDAREGULARMEMORYIMPORTIMAGECONTAINER_H__
