/*======================================================================================

	Program:		Range Imaging Toolkit (RITK)
	Module:			ritkVTKImageActorPipeline.h
	Version:		$Revision: 1821 $
	Date:			$Date: 2011-10-02 15:41:17 +0200 (So, 02 Okt 2011) $

	Copyright (c) 2010-2011
	Pattern Recognition Lab, 
    University Erlangen-Nuremberg, Germany
    All rights reserved.

	This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See RITKCopyright.txt for more information.

	http://www5.informatik.uni-erlangen.de/research/software/ritk

========================================================================================*/

#ifndef VTKIMAGEACTORPIPELINE_H__
#define VTKIMAGEACTORPIPELINE_H__

#include "ritkVTK.h"

#include <vtkMutexLock.h>
#include <vtkImageActor.h>
#include <vtkImageData.h>
#include <vtkSmartPointer.h>


namespace ritk
{


/**	@ingroup	ritkVTK
 *	@class		VTKImageActorPipeline
 *	@author		Sebastian Bauer
 *	@brief		Class that encapsulates a VTK image actor pipeline and ensures thread safe rendering
 *
 *	@details
 *	This class encapsulates a VTK image actor pipeline (i.e. ImageData->ImageActor) to be used in a VTK renderer.
 *	The purpose of this class is twofold: Firstly, it is provided for convenience to omit a manual set up 
 *	of the image actor pipeline that is the same for all image actors. As the VKTImageActorPipeline is derived from 
 *  vtkImageActor it can be passed to a VTK renderer directly. \n
 *	Secondly, an interface for thread safe rendering is provided. The need for a thread safe rendering interface 
 *	arises from the fact that the actor is eventually rendered in a QVTKWidget which comes along with temporal 
 *	arbitrary repaint events and, thus, temporal arbitrary access to the actor and its associated data that is 
 *	often managed by a different thread. \n
 *	Thread safe rendering is realized by deriving from vtkImageActor and overloading it's Render() method. The actual
 *	rendering is then locked by a mutex. The same mutex is used to lock the VTKImageActorPipeline's SetData() method that 
 *	takes a flag that indicates whether a deep copy of the data should be performed (data consistency is ensured by
 *	this class) or the caller takes full responsibility of data consistency. In addition, the VTKImageActorPipeline can be
 *	locked manually. \n
 *	Note: Tests revealed that additional methods are invoked during a repaint event, (e.g. GetBounds(), SetMapper()).
 *	All of them have to be identified and overloaded. In case of a crash please report the corresponding vtkActor() 
 *	method that eventually triggered the crash.
 */
class RITKVTK_EXPORT VTKImageActorPipeline : public vtkImageActor
{
public:
	/// Casting
	vtkTypeMacro(VTKImageActorPipeline,vtkImageActor);

	/// Object creation
	static VTKImageActorPipeline *New();

	/**	@name	Typedefs for the pipeline elements */
	//@{
	typedef vtkSmartPointer<vtkImageData>	VTKImageDataPointer;	
	//@}

	/**	@brief	Set the data
	 *	@param	Data		The data
	 *	@param	DeepCopy	Whether to perform a deep copy of the given data
	 *
	 *	@details
	 *	Will set the data for this actor pipeline. The VTKImageActorPipeline will be locked
	 *	during this operation. If the deep copy flag is set the data will be updated
	 *	before the copy operation.
	 *	@sa Render()
	 */
	virtual void SetData(VTKImageDataPointer Data, bool DeepCopy=true);

	/**	@brief	Set the input
	 *	@param	Data		The data to be used
	 *
	 *	@details
	 *	This method will set the input for the VTK image actor pipeline. The pipeline
	 *	will therefore be locked to ensure thread safety.
	 */
	virtual void SetInput(VTKImageDataPointer Data);

	/**	@brief	Overloaded rendering routine
	 *	@param	Renderer	Renderer
	 *	
	 *	@details
	 *	This method realizes a thread safe rendering routine. During rendering
	 *	the VKTImageActorPipeline is locked.
	 *	@sa SetData();
	 */
	virtual void Render(vtkRenderer *Renderer);

	/// Lock the VKTImageActorPipeline manually
	void Lock();

	/// Unlock the VKTImageActorPipeline manually
	void Unlock();

	/// Overloaded to enforce thread safety. Note: We should buffer the returned values!
	virtual double *GetBounds();

	/// Overloaded to enforce thread safety
	int RenderOpaqueGeometry(vtkViewport *viewport);

	/// Set the wrapped actor to modified
	void Modified();

protected:
	/// Constructor
	VTKImageActorPipeline();
	/// Destructor
	virtual ~VTKImageActorPipeline();

protected:
	/// Mutex used to lock the ActorPipeline.
	vtkSmartPointer<vtkMutexLock> m_Mutex;
	/// Device
	vtkSmartPointer<vtkImageActor> m_Device;
	/// The data	
	vtkSmartPointer<vtkImageData> m_Data;

private:
	VTKImageActorPipeline(const VTKImageActorPipeline&);  // Not implemented.
	void operator=(const VTKImageActorPipeline&);  // Not implemented.
};

} // end namespace ritk

#endif // VTKIMAGEACTORPIPELINE_H__